///////////////////////////////////////////////////////////////////////////////
//                                                                           //
//                         DXF File reader object/code                       //
//                             John Biddiscombe                             //
//                      Rutherford Appleton Laboratory, UK                   //
//                           j.biddiscombe@rl.ac.uk                          //
//                            DXF code release 2.1                           //
//                                                                           //
///////////////////////////////////////////////////////////////////////////////
//                                                                           //
// Thanks very much to John F Herbster for the original DXF class that got   //
// this started --- extract from his header follows...                       //
//                                                                           //
// Pgm. 07/14/95 by John F Herbster, CIS:72714,3445, Houston, TX.            //
// for Rick Rogers (CIS:74323,3573).                                         //
//                                                                           //
///////////////////////////////////////////////////////////////////////////////

unit DXF_read;

interface

uses
  { Borland }
  Windows,SysUtils,StdCtrls,ComCtrls,Dialogs,Classes,Graphics,
  { Mine }
  DXF_structs,DXF_Utils,Thinkbox;

// Thanks to Ian L. Kaplan, whose code contained these ID's
const
  DXF_start            = 0;
  DXF_textval          = 1;
  DXF_name             = 2;
  DXF_othername1       = 3;
  DXF_othername2       = 4;
  DXF_entity_handle    = 5;
  DXF_line_type        = 6;
  DXF_text_style       = 7;
  DXF_layer_name       = 8;
  DXF_var_name         = 9;
  DXF_primary_X        = 10;
  DXF_other_X_1        = 11;
  DXF_other_X_2        = 12;
  DXF_other_X_3        = 13;
  DXF_primary_Y        = 20;
  DXF_other_Y_1        = 21;
  DXF_other_Y_2        = 22;
  DXF_other_Y_3        = 23;
  DXF_primary_Z        = 30;
  DXF_other_Z_1        = 31;
  DXF_other_Z_2        = 32;
  DXF_other_Z_3        = 33;
  DXF_elevation        = 38;
  DXF_thickness        = 39;
  DXF_floatvals        = 40;
  DXF_AspectRatio      = 41;
  DXF_repeat           = 49;
  DXF_angle1           = 50;  DXF_angle2        = 51;
  DXF_angle3           = 52;  DXF_angle4        = 53;
  DXF_angle5           = 54;  DXF_angle6        = 55;
  DXF_angle7           = 56;  DXF_angle8        = 57;
  DXF_angle9           = 58;
  DXF_colornum         = 62;
  DXF_entities_flg     = 66;
  DXF_ent_ident        = 67;
  DXF_SeventyFlag      = 70;
  DXF_SeventyOneFlag   = 71;
  DXF_SeventyTwoFlag   = 72;
  DXF_view_state       = 69;
  DXF_comment          = 999;

///////////////////////////////////////////////////////////////////////////////
// DXF_Reader class definition
///////////////////////////////////////////////////////////////////////////////
Const
  MaxSizeOfBuf = 4096;

type
  tCharArray = array [0..MaxSizeOfBuf-1] of char;

type
  abstract_entity = class;

  DXF_Reader = class
  private
    IO_chan    : file;
    SizeOfBuf  : integer;
    num_in_buf : integer;
    ii         : integer;
    EC,fCode   : integer;
    pBuf       : ^tCharArray;
    Line_num   : longint;
    fLine      : shortstring;
    // Some DXF files have no layer info, so we will allow the automatic
    // creation of one layer named after the layer specified by the first entity
    // in the file
    special_no_layers_defined_flag : boolean;

    progress   : TProgressBar;
    // special additions to make parsing easier...
    file_pos   : integer;
    marked_pos : integer;
    backflag   : boolean;
    procedure   go_back_to_last(code:integer; str:shortstring);
    procedure   mark_position;
    procedure   goto_marked_position;
    // end of special additions
    procedure   go_back_to_start;
    function    NextGroupCode: integer;
    function    ValStr: shortstring;
    function    ValDbl: double;
    function    ValInt: integer;
    function    code_and_string(var group:integer; var s:string) : boolean;
    function    code_and_double(var group:integer; var d:double) : boolean;
    function    read_2Dpoint(var p1:Point3D)                     : boolean;
    function    skip_upto_section(name:string)                   : boolean;
    function    read_entity_data(ent:abstract_entity)            : boolean;
    function    read_generic(var layer:integer)                  : abstract_entity;
  public
    // Extents in (x,y) of the dataset
    min_extents    : Point3D;
    max_extents    : Point3D;
    // We will read the Entities in the layers into this list
    DXF_Layers     : TList;
    colour_BYLAYER : boolean;
    Aspect         : double;
    skipped        : TStrings;
    // Constructors and destructors
    Constructor Create (const aName: shortstring);
    Destructor  Destroy;                           override;
    // Header section
    function    move_to_header_section : boolean;
    function    read_header            : boolean;
    function    get_min_extent         : Point3D;
    function    get_max_extent         : Point3D;
    // Blocks section
    function    move_to_blocks_section : boolean;
    function    read_blocks            : boolean;
    // Tables section
    function    move_to_tables_section : boolean;
    function    read_tables : boolean;
    function    read_layer_information : boolean;
    function    read_vport_information : boolean;
    function    layer_num(layername:string) : integer;
    // Entities section
    function    move_to_entity_section : boolean;
    function    read_entities          : boolean;
    // we can read point,line,text,circle,arc with this one
    function    general_purpose_read(obj_type:TClass) : boolean;
    // entities that require a little more complexity
    function    read_insert   : boolean;
    function    read_polyline : boolean;
    // These are the main routines to use
    function    read_file                 : boolean;
    function    remove_empty_layers       : boolean;
    function    release_control_of_layers : TList;
    procedure   set_skipped_list(s:TStrings);
  end;

///////////////////////////////////////////////////////////////////////////////
// This is a simple class used only during file reads, it should not be used
// as a base for any objects. It is to simplify problems like...
// some files have 2D coords, some 3D, circles & arcs etc have extra group
// codes etc. etc. etc.
// Add extra group codes if you need to recognize them
///////////////////////////////////////////////////////////////////////////////
  abstract_entity = class
    p1,p2,p3,p4 : Point3D;
    radius_h    : double; // this val used for circle/arc radius and also text height
    angle1      : double;
    angle2      : double;
    colour      : integer;
    flag_70     : integer;
    attflag     : integer;
    s           : string;
    tagstr      : string;
    layer       : string;
    procedure setx1(d:double);
    procedure sety1(d:double);
    procedure setz1(d:double);
    procedure setx2(d:double);
    procedure sety2(d:double);
    procedure setz2(d:double);
    procedure setx3(d:double);
    procedure sety3(d:double);
    procedure setz3(d:double);
    procedure setx4(d:double);
    procedure sety4(d:double);
    procedure setz4(d:double);
    procedure setradius_or_height(d:double);
    procedure setang1(d:double);
    procedure setang2(d:double);
    procedure setcolour(c:integer);
    procedure setattflag(i:integer);
    procedure setflag_70(i:integer);
    procedure setstring(s_:string);
    procedure settag(s_:string);
    procedure setlayer(s_:string);
  end;

///////////////////////////////////////////////////////////////////////////////
// DXF File read exceptions will be this type
///////////////////////////////////////////////////////////////////////////////
type
  DXF_read_exception = class(Exception)
    line_number : integer;
    constructor create(err_msg:string; line:integer);
  end;

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
// implementation
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
implementation

///////////////////////////////////////////////////////////////////////////////
// DXFReader implementation
///////////////////////////////////////////////////////////////////////////////
Constructor DXF_Reader.Create (const aName: shortstring);
begin
  Inherited Create;
  AssignFile(IO_chan,aName);
  Reset(IO_chan,1);
  SizeOfBuf:=MaxSizeOfBuf;
  GetMem(pBuf,SizeOfBuf);
  DXF_Layers        := TList.Create;
  colour_BYLAYER    := false;
  Line_num          := 0;
  Aspect            := 1;
  backflag          := false;
  progress          := Thinking_box.bar;
  progress.position := 0;
  progress.max      := FileSize(IO_chan) div MaxSizeOfBuf;
  min_extents       := aPoint3D(0,0,0);
  max_extents       := aPoint3D(0,0,0);
end;

destructor DXF_Reader.Destroy;
var lp1 : integer;
begin
  if (DXF_Layers<>nil) then
    for lp1 := 0 to DXF_Layers.count-1 do DXF_Layer(DXF_Layers[lp1]).Free;
  DXF_Layers.Free;
  CloseFile(IO_chan);
  FreeMem(pBuf,SizeOfBuf);
  Inherited Destroy;
end;
{ --------------------------------------------------------------------------- }
{ Routines for fetching codes and values
{ --------------------------------------------------------------------------- }
procedure DXF_Reader.go_back_to_start;
begin
  Reset(IO_chan,1);
  num_in_buf := 0;
  ii         := 0;
end;

procedure DXF_Reader.go_back_to_last(code:integer; str:shortstring);
begin
  fCode    := code;
  fLine    := str;
  backflag := true;
end;

procedure DXF_Reader.mark_position;
begin
  marked_pos := File_pos + ii;
end;

procedure DXF_Reader.goto_marked_position;
begin
  Seek(IO_chan,marked_pos);
  File_pos := marked_pos;
  num_in_buf := 0;
  ii         := 0;
end;

function DXF_Reader.NextGroupCode: integer;
  function GotMore: boolean;
  begin
    file_pos := FilePos(IO_chan);
    BlockRead(IO_chan,pBuf^,SizeOfBuf,num_in_buf); ec:=IoResult; ii:=0;
    If (ec=0) and (num_in_buf=0) then ec:=-1; GotMore:=(ec=0);
    progress.position := progress.position+1;
  end{GotMore};

  function GotLine: boolean;
  const CR=#13; LF=#10;
  var c: char;
  begin
    byte(fLine[0]):=0;
    While (ii<num_in_buf) or GotMore do begin
      c:=pBuf^[ii]; inc(ii);
      If (c<>CR) and (c<>LF) and (length(fLine)<255) then begin
        inc(fLine[0]); fLine[length(fLine)]:=c
      end
      else begin      // Extra code added to handle C/Unix style LF not CR/LF
        if (c=CR) then begin
          if (ii<num_in_buf) or GotMore then begin
            if pBuf^[ii]=LF then begin inc(ii); break; end;
          end;
        end else if (c=LF) then break;
      end;
    end;
    GotLine:=(ec=0) and ((c=CR) or (c=LF));
    inc(Line_num);
  end;

begin {NextGroupCode}
  if backflag then begin
    result   := fCode;
    backflag := false;
  end
  else begin
    If GotLine then Val(fLine,fCode,ec);
    If ec<>0 then fCode:=-2
    else if not GotLine then fCode:=-2;
    Result:=fCode;
  end;
end {NextGroupCode};

function DXF_Reader.ValStr: shortstring;
begin Result:=fLine end;

function DXF_Reader.ValDbl: double;
begin
  Val(fLine,Result,ec);
  If ec<>0 then raise DXF_read_exception.Create('Invalid Floating point conversion',line_num);
end;

function DXF_Reader.ValInt: integer;
begin
  Val(fLine,Result,ec);
  If ec<>0 then raise DXF_read_exception.Create('Invalid Integer conversion',line_num);
end;

function DXF_Reader.code_and_string(var group:integer; var s:string) : boolean;
var astr : string;
begin
  result := true;
  group  := NextGroupCode;
  if group>=0 then s := ValStr
  else result := false;
  // useful in debugging
  //  if (group=0) then begin astr := IntToStr(group)+' '+s; alb.Items.Add(astr); end;
end;

function DXF_Reader.code_and_double(var group:integer; var d:double) : boolean;
begin
  result := true;
  group  := NextGroupCode;
  if group>=0 then d := Valdbl
  else result := false;
end;

// This routine is just for the $EXT(max/min) and should be used with care....
function DXF_Reader.read_2Dpoint(var p1:Point3D) : boolean;
var Groupcode : integer;
begin
  repeat Groupcode:=NextGroupCode;
  until (Groupcode=DXF_primary_X) or (Groupcode<0);
  if Groupcode<0 then begin result:=false; exit; end;
  p1.x := Valdbl;
  result := code_and_double(Groupcode,p1.y);  { y next              }
end;

function DXF_Reader.skip_upto_section(name:string) : boolean;
var Group   : integer;
    s       : string;
begin
  result := false;
  repeat
    if not code_and_string(Group,s) then break;
    if (Group=0) then begin
      if (s='SECTION') then begin
        if not code_and_string(Group,s) then break;
        if (Group=DXF_name) then begin
          if (s=name) then result := true
          else exit;
        end
        else if skipped<>nil then Skipped.Add(s);
      end else if skipped<>nil then Skipped.Add(s);
    end;
  until (result);
end;
{ --------------------------------------------------------------------------- }
{ Header section
{ --------------------------------------------------------------------------- }
function DXF_Reader.move_to_header_section : boolean;
begin
  result := skip_upto_section('HEADER');
end;

function DXF_Reader.read_header : boolean;
var Group : integer;
    s     : string;
begin
  result := false;
  repeat
    if not code_and_string(Group,s) then break;
    if (group=9) and (s='$EXTMAX') then begin
      if not read_2Dpoint(max_extents) then break;
    end;
    if (group=9) and (s='$EXTMIN') then begin
      if not read_2Dpoint(min_extents) then break;
    end;
    if (group=9) and (s='$CECOLOR') then begin
      if (NextGroupCode=DXF_colornum) and (ValInt=256) then colour_BYLAYER := true;
    end;
    result := (Group=0) and (s='ENDSEC');
  until result;
end;

function DXF_Reader.get_min_extent : Point3D;
begin
  result := min_extents;
end;

function DXF_Reader.get_max_extent : Point3D;
begin
  result := max_extents;
end;
{ --------------------------------------------------------------------------- }
{ Blocks section
{ --------------------------------------------------------------------------- }
function DXF_Reader.move_to_blocks_section : boolean;
begin
  result := skip_upto_section('BLOCKS');
end;

function DXF_Reader.read_blocks : boolean;
begin
  // I don't need any block info yet...
  result := true;
end;
{ --------------------------------------------------------------------------- }
{ Tables (Layers - VPort) section
{ --------------------------------------------------------------------------- }
function DXF_Reader.move_to_tables_section : boolean;
begin
  result := skip_upto_section('TABLES');
end;

function DXF_Reader.read_tables : boolean;
var Group : integer;
    s     : string;
begin
  result := false;
  repeat
    if not code_and_string(Group,s) then break;
    if (Group=0) and (s='TABLE') then begin
      if not code_and_string(Group,s) then break;
      if (Group=DXF_name) then begin
        if (s='LAYER') then read_layer_information
        else if (s='VPORT') then read_vport_information
        else if skipped<>nil then Skipped.Add(s);
      end;
    end;
    result := (Group=0) and (s='ENDSEC');
  until result;
end;

function DXF_Reader.read_layer_information : boolean;
var Group,Lay_num : integer;
    s             : string;
begin
  lay_num := -1;
  result  := false;
  repeat
    if not code_and_string(Group,s) then break;
    if (Group=0) then begin
      if (s='LAYER') then begin
        if not code_and_string(Group,s) then break;
        if (Group=DXF_name) then lay_num := DXF_Layers.Add(DXF_Layer.create(s));
      end
      else if (s='ENDTAB') then result := true
      else if skipped<>nil then Skipped.Add(s);
    end
    else if (Group=DXF_colornum) and (lay_num<>-1) then
      DXF_Layer(DXF_Layers[lay_num]).Colour := ValInt;
  until result;
end;

function DXF_Reader.read_vport_information : boolean;
var Group : integer;
    s     : string;
begin
  result := false;
  repeat
    if not code_and_string(Group,s) then break;
    if (Group=0) then begin
      if (s='VPORT') then begin
        if not code_and_string(Group,s) then break;
        if (Group=DXF_name) then begin
          if (s='*ACTIVE') then repeat
            if not code_and_string(Group,s) then break;
            if (Group=DXF_AspectRatio) then Aspect := ValDbl;
            result := (Group=0) and (s='ENDTAB');
          until (result)
          else if skipped<>nil then Skipped.Add(s);
        end;
      end
      else if skipped<>nil then Skipped.Add(s);
    end
  until (result);
end;

function DXF_Reader.layer_num(layername:string) : integer;
var lp1 : integer;
begin
  result := -1;
  for lp1:=0 to DXF_Layers.count-1 do begin
    if DXF_Layer(DXF_Layers[lp1]).name=layername then begin
      result := lp1;
      exit;
    end;
  end;
end;
{ --------------------------------------------------------------------------- }
{ Entities section
{ --------------------------------------------------------------------------- }
function DXF_Reader.move_to_entity_section : boolean;
begin
  result := skip_upto_section('ENTITIES');
end;

function DXF_Reader.read_entities : boolean;
var Groupcode : integer;
    s         : string;
begin
  result := false;
  repeat
    try
      if not code_and_string(Groupcode,s) then break;
      if (Groupcode=0) then begin
        if (s='POINT') then begin if not general_purpose_read(Point_) then
          raise DXF_read_exception.Create('Error reading Point entity',line_num); end
        else if (s='INSERT') then begin if not read_insert then
          raise DXF_read_exception.Create('Error reading Insert entity',line_num); end
        else if (s='TEXT') then begin if not general_purpose_read(Text_) then
          raise DXF_read_exception.Create('Error reading Text entity',line_num); end
        else if (s='LINE') then begin if not general_purpose_read(Line_) then
          raise DXF_read_exception.Create('Error reading Line entity',line_num); end
        else if (s='POLYLINE') then begin if not read_polyline then
          raise DXF_read_exception.Create('Error reading Polyline entity',line_num); end
        else if (s='3DFACE') then begin if not general_purpose_read(Face3D_) then
          raise DXF_read_exception.Create('Error reading 3DFace entity',line_num); end
        else if (s='CIRCLE') then begin if not general_purpose_read(Circle_) then
          raise DXF_read_exception.Create('Error reading Circle entity',line_num); end
        else if (s='ARC') then begin if not general_purpose_read(Arc_) then
          raise DXF_read_exception.Create('Error reading Arc entity',line_num); end
        else if (s='ENDSEC') then result := true
        else if skipped<>nil then Skipped.Add(s);
      end;
    except
      on E:DXF_read_exception do begin
        stopped_thinking;
        if MessageBox(0,@E.message[1],'DXF read error warning',MB_OKCANCEL)=IDCANCEL then
          raise DXF_read_exception.Create('User aborted',-1);
        thinking_bar(nil,'Reading DXF file...');
      end;
      on E:Exception do Showmessage(E.Message);
    end;
  until result;
end;
{ --------------------------------------------------------------------------- }
{ Entity reading code
{ --------------------------------------------------------------------------- }
function DXF_Reader.read_entity_data(ent:abstract_entity) : boolean;
var Groupcode : integer;
begin
  repeat
    Groupcode := NextGroupCode;
    case Groupcode of
      DXF_layer_name   : ent.setlayer(ValStr);
      DXF_primary_X    : ent.setx1(Valdbl);
      DXF_primary_Y    : ent.sety1(Valdbl);
      DXF_primary_Z    : ent.setz1(Valdbl);
      DXF_floatvals    : ent.setradius_or_height(Valdbl);
      DXF_angle1       : ent.setang1(Valdbl);
      DXF_angle2       : ent.setang2(Valdbl);
      DXF_other_X_1    : ent.setx2(Valdbl);
      DXF_other_Y_1    : ent.sety2(Valdbl);
      DXF_other_Z_1    : ent.setz2(Valdbl);
      DXF_other_X_2    : ent.setx3(Valdbl);
      DXF_other_Y_2    : ent.sety3(Valdbl);
      DXF_other_Z_2    : ent.setz3(Valdbl);
      DXF_other_X_3    : ent.setx4(Valdbl);
      DXF_other_Y_3    : ent.sety4(Valdbl);
      DXF_other_Z_3    : ent.setz4(Valdbl);
      DXF_SeventyFlag  : ent.setflag_70(ValInt);
      DXF_colornum     : ent.setcolour(ValInt);
      DXF_entities_flg : ent.setattflag(ValInt);
      DXF_textval      : ent.setstring(ValStr);
      DXF_name         : ent.settag(ValStr);
    end;
  until (Groupcode<=0); // end or fault;
  if Groupcode<0 then begin result:=false; exit; end;
  // we need to put the code=0, and valstr back, so the next entity starts
  // with the zero when neccessary
  go_back_to_last(Groupcode,fline);
  result := true;
end;

function DXF_Reader.read_generic(var layer:integer) : abstract_entity;
var ent : abstract_entity;
    s   : string;
begin
  result := nil;
  ent := abstract_entity.create;
  if read_entity_data(ent) then begin
    layer := layer_num(ent.layer);
    if layer>=0 then begin
      result := ent;
      exit;
    end else begin
      if special_no_layers_defined_flag then begin
        layer := DXF_Layers.Add(DXF_Layer.create(ent.layer));
        special_no_layers_defined_flag := false;
        result := ent;
        exit;
      end
      else begin
        s := ent.layer;
        ent.Free;
        raise DXF_read_exception.Create('Invalid Layer ('+s+')',line_num);
      end;
    end;
  end;
end;

{ These ones are simple, so we use a crafty TClass parameter }
function DXF_Reader.general_purpose_read(obj_type:TClass) : boolean;
var ent   : abstract_entity;
    inst  : DXF_Entity;
    layer : integer;
begin
  ent := read_generic(layer);
  if ent<>nil then begin
    with ent do begin
      if      obj_type = Point_  then inst := Point_. create(p1,colour)
      else if obj_type = Text_   then inst := Text_.  create(p1,s,radius_h,colour)
      else if obj_type = Line_   then inst := Line_.  create(p1,p2,colour)
      else if obj_type = Circle_ then inst := Circle_.create(p1,radius_h,colour)
      else if obj_type = Arc_    then inst := Arc_.   create(p1,radius_h,angle1,angle2,colour)
      else if obj_type = Face3D_ then begin
        if p1_eq_p2_3d(p3,p4) then inst := Face3D_.create(3, @p1, colour,true)
        else                       inst := Face3D_.create(4, @p1, colour,true)
      end;
    end;
    DXF_Layer(DXF_Layers[layer]).add_entity_to_layer(inst);
    ent.Free;
    result := true;
  end;
end;

{ INSERTs may have ATTRIBs which makes it a little more complicated }
function DXF_Reader.read_insert : boolean;
var ent,ent2 : abstract_entity;
    code,num : integer;
    layer    : integer;
    atts     : array[0..255] of Attrib_;
begin
  result := true;
  num := 0;
  ent := read_generic(layer);
  if ent<>nil then begin
    if ent.attflag=1 then begin
      repeat
        result := (Nextgroupcode=0);
        if result and (ValStr='ATTRIB') then begin
          ent2 := read_generic(layer);
          if ent2<>nil then begin
            atts[num] := Attrib_.create(ent2.p1,ent2.s,ent2.tagstr,ent2.radius_h,ent2.colour);
            ent2.Free;
            inc(num);
          end else result := false;
        end;
      until (not result) or (ValStr='SEQEND');
      if result then Nextgroupcode; // remove the SEQEND put back
    end;
    DXF_Layer(DXF_Layers[layer]).add_entity_to_layer(
      Insert_.create(ent.p1,ent.colour,num,@atts[0]));
    ent.Free;
  end
  else result := false;
end;

{ POLYLINEs have variable number of points... }
function DXF_Reader.read_polyline : boolean;
var ent1,ent2    : abstract_entity;
    vertices,lp1 : integer;
    exceeded     : boolean;
    v1,v2        : integer;
    tempvert     : array[0..1023] of Point3D;
    layercode    : integer;
    closed_poly  : boolean;
    s            : string;
begin
  result := false; exceeded := false;
  closed_poly := false;
  ent1   := abstract_entity.create;
  // read initial polyline data
  if not read_entity_data(ent1) then begin ent1.Free; exit; end;
  layercode := layer_num(ent1.layer);
  if layercode=-1 then begin
    if special_no_layers_defined_flag then begin
      layercode := DXF_Layers.Add(DXF_Layer.create(ent1.layer));
      special_no_layers_defined_flag := false;
    end
    else begin
      s := ent1.layer;
      ent1.Free;
      raise DXF_read_exception.Create('Invalid Layer ('+s+')',line_num);
    end;
  end;
  vertices  := 0;
  // read each vertex
  ent2   := abstract_entity.create;
  repeat
    if (NextGroupCode=0) and (ValStr = 'VERTEX') then begin
      if read_entity_data(ent2) then begin
        tempvert[vertices] := ent2.p1;
        inc(vertices);
      end else begin ent1.Free; ent2.Free; exit; end;
    end;
  until fLine='SEQEND';
  // this should set result to true, because 0 SEQEND is next
  result := NextGroupCode=0;
  if ((ent1.Flag_70) and 1)=1 then closed_poly := true;
  if vertices<=max_vertices_per_polyline then begin
    DXF_Layer(DXF_Layers[layercode]).add_entity_to_layer(
      Polyline_.create(vertices,@tempvert[0],ent1.colour,closed_poly));
  end else begin
    closed_poly := false;
    exceeded   := true;
    v1         := 0;
    v2         := vertices;
    while v2>max_vertices_per_polyline do begin
      DXF_Layer(DXF_Layers[layercode]).add_entity_to_layer(
        Polyline_.create(max_vertices_per_polyline,@tempvert[v1],ent1.colour,closed_poly));
      v2 := v2 - max_vertices_per_polyline;
      v1 := v1 + max_vertices_per_polyline;
    end;
    DXF_Layer(DXF_Layers[layercode]).add_entity_to_layer(
      Polyline_.create(v2,@tempvert[v1],ent1.colour,closed_poly));
  end;
  ent1.Free; ent2.Free;
  if exceeded then
    raise DXF_read_exception.Create('Polyline contained '+IntToStr(vertices)+
      ' points : Broken into multiple segments',line_num);
end;
///////////////////////////////////////////////////////////////////////////////
// Main routines to use
///////////////////////////////////////////////////////////////////////////////
function DXF_Reader.read_file : boolean;
var lp1 : integer;
begin
  result := true;
  thinking_bar(nil,'Reading DXF file...');
  special_no_layers_defined_flag := false;
  try
    mark_position;
    if not (move_to_header_section and read_header) then begin
      Thinking(nil,'No Header or invalid Header section in DXF file');
      Sleep(1500);
      goto_marked_position;
    end;
    mark_position;
    if not (move_to_tables_section and read_tables) then begin
      Thinking(nil,'No Layers or invalid Tables section in DXF file');
      Sleep(1500);
      goto_marked_position;
      special_no_layers_defined_flag := true;
    end;
    mark_position;
    if not (move_to_blocks_section and read_blocks) then goto_marked_position;

    thinking_bar(nil,'Reading DXF file...');
    if not (move_to_entity_section and read_entities) then
      raise DXF_read_exception.Create('No Entities or invalid Entities section in DXF file',-1);
  except
    on E:DXF_read_exception do begin
      stopped_thinking;
      MessageBox(0,@E.message[1],'DXF Read Error',MB_ICONWARNING);
    end;
    on E:EAccessViolation do begin
      stopped_thinking;
      MessageDlg(E.message, mtWarning, [mbOK], 0);
    end;
  end;
  if p1_eq_p2_3D(min_extents,aPoint3D(0,0,0)) or p1_eq_p2_3D(max_extents,aPoint3D(0,0,0)) then begin
    thinking(nil,'File contained no Max/Min extents. Scanning...');
    sleep(1500); // just a delay to let the message be visible
    for lp1:=0 to DXF_layers.count-1 do
      DXF_Layer(DXF_Layers[lp1]).max_min_extents(max_extents,min_extents);
  end;
  stopped_thinking;
end;

function DXF_Reader.remove_empty_layers : boolean;
var lp1   : integer;
    layer : DXF_layer;
begin
   for lp1 := DXF_Layers.count-1 downto 0 do begin
     layer :=  DXF_Layers[lp1];
     if layer.num_lists=0 then begin
       DXF_Layers.Remove(layer);
       layer.Free;
     end;
  end;
  result := (DXF_Layers.count>0);
end;

// Hand over ownership of the layers, the owner of the entity lists
// is now responsible for their destruction
function DXF_Reader.release_control_of_layers : TList;
begin
  result     := DXF_Layers;
  DXF_Layers := nil;
end;

// Since we're not reading all groupcodes, we offer the chance
// to dump the main titles into a list so we can see what
// we've missed
procedure DXF_Reader.set_skipped_list(s:TStrings);
begin
  skipped := s;
end;
///////////////////////////////////////////////////////////////////////////////
// abstract_entity class member functions
///////////////////////////////////////////////////////////////////////////////
procedure abstract_entity.setx1(d:double);
begin p1.x := d; end;
procedure abstract_entity.sety1(d:double);
begin p1.y := d; end;
procedure abstract_entity.setz1(d:double);
begin p1.z := d; end;
procedure abstract_entity.setx2(d:double);
begin p2.x := d; end;
procedure abstract_entity.sety2(d:double);
begin p2.y := d; end;
procedure abstract_entity.setz2(d:double);
begin p2.z := d; end;
procedure abstract_entity.setx3(d:double);
begin p3.x := d; end;
procedure abstract_entity.sety3(d:double);
begin p3.y := d; end;
procedure abstract_entity.setz3(d:double);
begin p3.z := d; end;
procedure abstract_entity.setx4(d:double);
begin p4.x := d; end;
procedure abstract_entity.sety4(d:double);
begin p4.y := d; end;
procedure abstract_entity.setz4(d:double);
begin p4.z := d; end;
procedure abstract_entity.setradius_or_height(d:double);
begin radius_h := d; end;
procedure abstract_entity.setang1(d:double);
begin angle1 := d; end;
procedure abstract_entity.setang2(d:double);
begin angle2 := d; end;
procedure abstract_entity.setcolour(c:integer);
begin colour := c; end;
procedure abstract_entity.setattflag(i:integer);
begin attflag := i; end;
procedure abstract_entity.setflag_70(i:integer);
begin flag_70 := i; end;
procedure abstract_entity.setstring(s_:string);
begin s := s_; end;
procedure abstract_entity.settag(s_:string);
begin tagstr := s_; end;
procedure abstract_entity.setlayer(s_:string);
begin layer := s_; end;
///////////////////////////////////////////////////////////////////////////////
// DXF File exception
///////////////////////////////////////////////////////////////////////////////
constructor DXF_read_exception.create(err_msg:string; line:integer);
begin
  if line>-1 then
    message := err_msg + #13#10 + 'Error occured at or near line number ' + IntToStr(line)
  else message := err_msg;
end;

initialization
end.


